#
# DeepLabCut Toolbox (deeplabcut.org)
# © A. & M.W. Mathis Labs
# https://github.com/DeepLabCut/DeepLabCut
#
# Please see AUTHORS for contributors.
# https://github.com/DeepLabCut/DeepLabCut/blob/main/AUTHORS
#
# Licensed under GNU Lesser General Public License v3.0
#

# Script for reproducing results in Zhou* & Stoffl* et al. for BUCTD with CoAM

# path=datapath
# results=resultspath or put numbers

# train model

# evaluate and
# check if predicted is close to result
