#
# DeepLabCut Toolbox (deeplabcut.org)
# © A. & M.W. Mathis Labs
# https://github.com/DeepLabCut/DeepLabCut
#
# Please see AUTHORS for contributors.
# https://github.com/DeepLabCut/DeepLabCut/blob/master/AUTHORS
#
# Licensed under GNU Lesser General Public License v3.0
#
from deeplabcut.gui.tabs.analyze_videos import AnalyzeVideos
from deeplabcut.gui.tabs.create_project import ProjectCreator
from deeplabcut.gui.tabs.create_training_dataset import CreateTrainingDataset
from deeplabcut.gui.tabs.create_videos import CreateVideos
from deeplabcut.gui.tabs.evaluate_network import EvaluateNetwork
from deeplabcut.gui.tabs.extract_frames import ExtractFrames
from deeplabcut.gui.tabs.extract_outlier_frames import ExtractOutlierFrames
from deeplabcut.gui.tabs.label_frames import LabelFrames
from deeplabcut.gui.tabs.manage_project import ManageProject
from deeplabcut.gui.tabs.modelzoo import ModelZoo
from deeplabcut.gui.tabs.open_project import OpenProject
from deeplabcut.gui.tabs.refine_tracklets import RefineTracklets
from deeplabcut.gui.tabs.train_network import TrainNetwork
from deeplabcut.gui.tabs.unsupervised_id_tracking import UnsupervizedIdTracking
from deeplabcut.gui.tabs.video_editor import VideoEditor
