#
# DeepLabCut Toolbox (deeplabcut.org)
# © A. & M.W. Mathis Labs
# https://github.com/DeepLabCut/DeepLabCut
#
# Please see AUTHORS for contributors.
# https://github.com/DeepLabCut/DeepLabCut/blob/main/AUTHORS
#
# Licensed under GNU Lesser General Public License v3.0
#
BASE_URL = "https://deeplabcut.github.io/DeepLabCut/docs/"
README = "https://deeplabcut.github.io/DeepLabCut/README.html"
URL_3D = BASE_URL + "Overviewof3D.html"
URL_MA_CONFIGURE = BASE_URL + "maDLC_UserGuide.html#configure-the-project"
URL_USE_GUIDE_SCENARIO = BASE_URL + "UseOverviewGuide.html#what-scenario-do-you-have"
